"""s3path boto3 utilities."""
import concurrent.futures
import logging
import os
import shutil
from pathlib import Path
from typing import Dict, Optional, Union

import boto3
from s3path import S3Path

from wpdatautil.pathlib import path_to_uri
from wpdatautil.timeit import Timer

AnyPath = Union[Path, S3Path]

log = logging.getLogger(__name__)


def cp_file(input_path: AnyPath, output_path: AnyPath, *, boto3_kwargs: Optional[Dict] = None) -> None:
    """Copy file from local or S3 path to local or S3 path.

    The parent directory of the output path is created if it doesn't exist.

    `boto3_kwargs` are forwarded to the respective call if boto3 is used.
    """
    boto3_kwargs = boto3_kwargs if (boto3_kwargs is not None) else {}
    output_path.parent.mkdir(parents=True, exist_ok=True)
    if (not isinstance(input_path, S3Path)) and (not isinstance(output_path, S3Path)):
        shutil.copy(input_path, output_path)
    elif isinstance(input_path, S3Path) and isinstance(output_path, S3Path):
        # Note: boto3.client('s3').copy_object works only for objects up to 5G in size.
        source = {"Bucket": input_path.bucket.name, "Key": str(input_path.key)}
        boto3.resource("s3").meta.client.copy(CopySource=source, Bucket=output_path.bucket.name, Key=str(output_path.key), **boto3_kwargs)
    elif (not isinstance(input_path, S3Path)) and isinstance(output_path, S3Path):
        boto3.resource("s3").meta.client.upload_file(Filename=str(input_path), Bucket=output_path.bucket.name, Key=str(output_path.key), **boto3_kwargs)
    elif isinstance(input_path, S3Path) and (not isinstance(output_path, S3Path)):
        boto3.resource("s3").meta.client.download_file(Bucket=input_path.bucket.name, Key=str(input_path.key), Filename=str(output_path), **boto3_kwargs)
    else:
        assert False


def cp_files(input_path: AnyPath, output_path: AnyPath, *, glob: str = "**/*", thread_pool_size: Optional[int] = os.cpu_count()) -> None:
    """Recursively copy files concurrently from glob pattern applied to the input local or S3 path to the output local or S3 path."""
    description = f"{glob} files from {path_to_uri(input_path)} to {path_to_uri(output_path)}"
    description += f" using a pool of {thread_pool_size} threads" if thread_pool_size else ""
    log.info(f"Copying {description}.")
    timer = Timer()
    futures = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=thread_pool_size) as pool:
        for input_file_path in input_path.glob(glob):
            if not input_file_path.is_file():
                continue
            output_file_path = output_path / input_file_path.relative_to(input_path)
            future = pool.submit(cp_file, input_file_path, output_file_path)
            futures.append(future)
        for future in concurrent.futures.as_completed(futures):
            future.result()
    log.info(f"Copied {len(futures):,} {description} in {timer}.")



"""timeit utilities."""
import datetime
import timeit


class Timer:
    """Measure time used."""

    # Ref: https://stackoverflow.com/a/57931660/
    def __init__(self, round_n_digits: int = 0):
        self._round_n_digits = round_n_digits
        self._start_time = timeit.default_timer()

    def __call__(self) -> float:
        return timeit.default_timer() - self._start_time

    def __str__(self) -> str:
        return str(datetime.timedelta(seconds=round(self(), self._round_n_digits)))



"""str SQL utilities."""
import textwrap
from typing import Dict, List, Optional


def sql_column_selections(inputs: List[str], transformations: Dict[str, Optional[str]]) -> Dict[str, str]:
    """Return a mapping of SQL SELECT clauses given the input columns after applying the given transformations.

    Example: column_selections(['c1', 'c2', 'c3'], transformations=['c2': None, 'c3': '{col} * 2']) -> {'c1': 'c1', 'c3': 'c3 * 2'}

    :param inputs: Input column names.
    :param transformations: Applicable transformations. If an input column is to be transformed, its name is the key and its value is the transformation.
                            In the value, {col} is replaced with the input column name.
    If the value is None, the column is skipped.
    """
    for column_name in transformations:
        assert column_name in inputs, {"column_name": column_name, "inputs": inputs, "transformations": transformations}

    selections = {}
    for column_name in inputs:
        if column_name in transformations:
            transformed_column = transformations[column_name]
            if transformed_column is None:
                continue  # Skip column.
            transformed_column = transformed_column.format(col=column_name)
            selections[column_name] = transformed_column
        else:
            selections[column_name] = column_name

    return selections


def sql_column_selections_str(inputs: Dict[str, str]) -> str:
    """Return a string representation of a mapping of SQL SELECT clauses.

    Example: {'c1': 'c1', 'c2': 'c2 * 2'} -> 'c1, c2 * 2 AS c2'
    """
    selections = ", ".join(k if k == v else f"\n{v} AS {k}" for k, v in inputs.items())
    if "\n" in selections:
        selections = "\n" + textwrap.indent(selections, prefix=" " * 4 * 4)
    return selections



"""humanfriendly utilities."""
import humanfriendly


def format_size(num_bytes: int) -> str:
    """Return the given number of bytes formatted as a human-readable size string using binary multiples.

    Examples
    --------
    0 -> "0 bytes"
    123456789 -> "117.74 MiB"
    3221225472 -> "3 GiB"

    """
    return humanfriendly.format_size(num_bytes, binary=True)



"""s3path utilities."""
import logging
import shutil
from pathlib import Path
from typing import Union

from s3path import S3Path

from wpdatautil.pathlib import path_to_uri

AnyPath = Union[Path, S3Path]

log = logging.getLogger(__name__)


def rmdirtree(path: AnyPath) -> None:
    """Remove the local or S3 path."""
    uri = path_to_uri(path)
    log.info(f"Removing path {uri}.")
    if isinstance(path, S3Path):
        if path.exists():
            path.rmdir()
    else:
        shutil.rmtree(path, ignore_errors=True)
    assert list(path.rglob("*")) == []
    log.info(f"Removed path {uri}.")
